#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

require('babel-polyfill');
var yargs = _interopDefault(require('yargs'));
var fs = _interopDefault(require('fs'));
var zlib = _interopDefault(require('zlib'));
var _ = _interopDefault(require('lodash'));
var fse = _interopDefault(require('fs-extra'));
var rimraf = _interopDefault(require('rimraf'));
var crypto = _interopDefault(require('crypto'));
var humanize = _interopDefault(require('string-humanize'));
var retrieveArguments = _interopDefault(require('retrieve-arguments'));

var asyncGenerator = function () {
  function AwaitValue(value) {
    this.value = value;
  }

  function AsyncGenerator(gen) {
    var front, back;

    function send(key, arg) {
      return new Promise(function (resolve, reject) {
        var request = {
          key: key,
          arg: arg,
          resolve: resolve,
          reject: reject,
          next: null
        };

        if (back) {
          back = back.next = request;
        } else {
          front = back = request;
          resume(key, arg);
        }
      });
    }

    function resume(key, arg) {
      try {
        var result = gen[key](arg);
        var value = result.value;

        if (value instanceof AwaitValue) {
          Promise.resolve(value.value).then(function (arg) {
            resume("next", arg);
          }, function (arg) {
            resume("throw", arg);
          });
        } else {
          settle(result.done ? "return" : "normal", result.value);
        }
      } catch (err) {
        settle("throw", err);
      }
    }

    function settle(type, value) {
      switch (type) {
        case "return":
          front.resolve({
            value: value,
            done: true
          });
          break;

        case "throw":
          front.reject(value);
          break;

        default:
          front.resolve({
            value: value,
            done: false
          });
          break;
      }

      front = front.next;

      if (front) {
        resume(front.key, front.arg);
      } else {
        back = null;
      }
    }

    this._invoke = send;

    if (typeof gen.return !== "function") {
      this.return = undefined;
    }
  }

  if (typeof Symbol === "function" && Symbol.asyncIterator) {
    AsyncGenerator.prototype[Symbol.asyncIterator] = function () {
      return this;
    };
  }

  AsyncGenerator.prototype.next = function (arg) {
    return this._invoke("next", arg);
  };

  AsyncGenerator.prototype.throw = function (arg) {
    return this._invoke("throw", arg);
  };

  AsyncGenerator.prototype.return = function (arg) {
    return this._invoke("return", arg);
  };

  return {
    wrap: function (fn) {
      return function () {
        return new AsyncGenerator(fn.apply(this, arguments));
      };
    },
    await: function (value) {
      return new AwaitValue(value);
    }
  };
}();



var asyncToGenerator = function (fn) {
  return function () {
    var gen = fn.apply(this, arguments);
    return new Promise(function (resolve, reject) {
      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          return Promise.resolve(value).then(function (value) {
            step("next", value);
          }, function (err) {
            step("throw", err);
          });
        }
      }

      return step("next");
    });
  };
};

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();







var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};





















var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();











var toArray = function (arr) {
  return Array.isArray(arr) ? arr : Array.from(arr);
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var Stage = function () {
    function Stage() {
        classCallCheck(this, Stage);
    }

    createClass(Stage, [{
        key: "and",
        value: function and() {
            return this;
        }
    }, {
        key: "but",
        value: function but() {
            return this;
        }
    }, {
        key: "with",
        value: function _with() {
            return this;
        }
    }, {
        key: "given",
        value: function given() {
            return this;
        }
    }, {
        key: "when",
        value: function when() {
            return this;
        }
    }, {
        key: "then",
        value: function then() {
            return this;
        }
    }]);
    return Stage;
}();

var StageMetadataStoreImpl = function () {
    function StageMetadataStoreImpl() {
        classCallCheck(this, StageMetadataStoreImpl);
        this.properties = [];
    }

    createClass(StageMetadataStoreImpl, [{
        key: 'addProperty',
        value: function addProperty(property) {
            this.properties.push(property);
        }
    }, {
        key: 'getProperties',
        value: function getProperties() {
            return this.properties;
        }
    }]);
    return StageMetadataStoreImpl;
}();

var REPORTS_DESTINATION = '.jsGiven-reports';

var INTRO_WORD_METHODS = ['given', 'when', 'then', 'and', 'but', 'with'];

var ScenarioPart = function () {
    function ScenarioPart(kind) {
        var steps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
        classCallCheck(this, ScenarioPart);

        this.kind = kind;
        this.steps = steps;
        this.introWord = null;
    }

    createClass(ScenarioPart, [{
        key: 'stageMethodCalled',
        value: function stageMethodCalled(methodName, parameters, stepStatus, executionTimeInNanos) {
            if (INTRO_WORD_METHODS.find(function (introWord) {
                return introWord === methodName;
            })) {
                this.introWord = methodName;
            } else {
                var isFirstStep = this.steps.length === 0;
                this.steps.push(new Step(methodName, parameters, isFirstStep, this.introWord, stepStatus, executionTimeInNanos));
                this.introWord = null;
            }
        }
    }]);
    return ScenarioPart;
}();

var Step = function Step(methodName, parameters, isFirstStep, introWord, stepStatus, durationInNanos) {
    classCallCheck(this, Step);

    var TWO_DOLLAR_PLACEHOLDER = 'zzblablaescapedollarsignplaceholdertpolm';

    var parametersCopy = [].concat(toConsumableArray(parameters));

    var words = [].concat(toConsumableArray(methodName // 'a_bill_of_$_$$'
    .replace('$$', TWO_DOLLAR_PLACEHOLDER) // 'a_bill_of_$_TWO_DOLLAR_PLACEHOLDER'
    .split('$') // ['a_bill_of', 'TWO_DOLLAR_PLACEHOLDER']
    .map(function (word) {
        return _.lowerCase(humanize(word));
    }) //  ['a bill of', 'TWO_DOLLAR_PLACEHOLDER']
    .reduce(function (previous, newString, index) {
        if (index === 0) {
            return [{ word: newString, scenarioParameterName: null }];
        }

        var formattedParameters = void 0;
        if (parametersCopy.length > 0) {
            var _parametersCopy$splic = parametersCopy.splice(0, 1),
                _parametersCopy$splic2 = slicedToArray(_parametersCopy$splic, 1),
                parameter = _parametersCopy$splic2[0];

            var _word = formatParameter(parameter.value, parameter.formatters);
            formattedParameters = [{
                word: _word,
                scenarioParameterName: parameter.scenarioParameterName
            }];
        } else {
            formattedParameters = [];
        }

        return [].concat(toConsumableArray(previous), toConsumableArray(formattedParameters), [{ word: newString, scenarioParameterName: null }]);
    }, []) //  ['a bill of', '500', 'TWO_DOLLAR_PLACEHOLDER']
    .filter(function (_ref) {
        var word = _ref.word;
        return word !== '';
    }) // If one puts a $ at the end of the method, this adds a useless '' at the end
    .map(function (_ref2) {
        var word = _ref2.word,
            scenarioParameterName = _ref2.scenarioParameterName;
        return {
            word: word.replace(TWO_DOLLAR_PLACEHOLDER, '$'),
            scenarioParameterName
        };
    }) //  ['a bill of', '500', '$']
    .map(toWord)), toConsumableArray(parametersCopy.map(function (parameter) {
        return {
            word: formatParameter(parameter.value, parameter.formatters),
            scenarioParameterName: parameter.scenarioParameterName
        };
    }).map(toWord)));

    if (introWord !== null) {
        words = [toIntroWord(introWord)].concat(toConsumableArray(words));
    }

    if (isFirstStep) {
        var _words = words,
            _words2 = toArray(_words),
            _words2$ = _words2[0],
            value = _words2$.value,
            isIntroWord = _words2$.isIntroWord,
            rest = _words2.slice(1);

        words = [new Word(_.upperFirst(value), isIntroWord)].concat(toConsumableArray(rest));
    }
    this.words = words;
    this.name = words.map(function (_ref3) {
        var value = _ref3.value;
        return value;
    }).join(' ');
    this.status = stepStatus;
    this.durationInNanos = durationInNanos;

    function toWord(_ref4) {
        var word = _ref4.word,
            scenarioParameterName = _ref4.scenarioParameterName;

        return new Word(word, false, scenarioParameterName);
    }

    function toIntroWord(value) {
        return new Word(value, true, null);
    }
};

var Word = function Word(value, isIntroWord) {
    var scenarioParameterName = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    classCallCheck(this, Word);

    this.value = value;
    this.isIntroWord = isIntroWord;
    this.scenarioParameterName = scenarioParameterName;
};



var ScenarioReport = function () {
    function ScenarioReport(groupReport, name, cases, argumentNames) {
        classCallCheck(this, ScenarioReport);

        this.groupReport = groupReport;
        this.name = name;
        this.cases = cases;
        this.argumentNames = argumentNames;
    }

    createClass(ScenarioReport, [{
        key: 'dumpToFile',
        value: function dumpToFile(reportsDestination) {
            createDirOrDoNothingIfExists(reportsDestination);
            var fileName = computeScenarioFileName(this.groupReport.name, this.name);
            fs.writeFileSync(`${reportsDestination}/${fileName}`, JSON.stringify(this), 'utf-8');
        }
    }]);
    return ScenarioReport;
}();

function formatParameter(parameter, formatters) {
    if (formatters.length > 0) {
        var value = parameter;
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = formatters[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var formatter = _step.value;

                value = formatter(value);
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        return formatParameter(value, []);
    }
    if (_.isString(parameter)) {
        return parameter;
    } else if (_.isObject(parameter) || Array.isArray(parameter)) {
        if (parameter.toString && parameter.toString !== Object.prototype.toString && parameter.toString !== Array.prototype.toString) {
            return parameter.toString();
        }
        return JSON.stringify(parameter);
    } else {
        return parameter && parameter.toString ? parameter.toString() : JSON.stringify(parameter);
    }
}



function createDirOrDoNothingIfExists(path) {
    try {
        fs.mkdirSync(path);
    } catch (error) {
        if (error.code !== 'EEXIST') {
            throw error;
        } else {
            // do nothing
        }
    }
}

function computeScenarioFileName(groupName, scenarioName) {
    var hash = crypto.createHash('sha256');
    hash.update(groupName + '\n' + scenarioName);
    return hash.digest('hex');
}

var generateJGivenReport = function () {
    var _ref = asyncToGenerator(function* (fail) {
        try {
            if (!directoryExists(`./${REPORTS_DESTINATION}`)) {
                console.log('No jsGiven reports found, skipping jgiven report generation');
                return;
            }

            yield installJGivenReportApp();
            generateJGivenReportDataFiles();

            console.log(`jsGiven report available in ./jGiven-report`);
        } catch (error) {
            console.log(error);
            process.exit(-1);
        }

        if (fail) {
            process.exit(1);
        }
    });

    return function generateJGivenReport(_x) {
        return _ref.apply(this, arguments);
    };
}();

var cleanReports = function () {
    var _ref2 = asyncToGenerator(function* () {
        try {
            yield removeDir(`./${REPORTS_DESTINATION}`);
        } catch (error) {
            console.log(error);
            process.exit(-1);
        }
    });

    return function cleanReports() {
        return _ref2.apply(this, arguments);
    };
}();

var installJGivenReportApp = function () {
    var _ref3 = asyncToGenerator(function* () {
        var reportPrefix = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '.';

        var reportDir = `${reportPrefix}/jGiven-report`;

        yield removeDir(reportDir);
        fse.copySync('node_modules/jgiven-html-app/dist/', reportDir);
        fs.mkdirSync(`${reportDir}/data`);
    });

    return function installJGivenReportApp() {
        return _ref3.apply(this, arguments);
    };
}();

function directoryExists(dirName) {
    try {
        return fs.statSync(dirName).isDirectory();
    } catch (error) {
        return false;
    }
}

function generateJGivenReportDataFiles() {
    var filter = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : function () {
        return true;
    };
    var reportPrefix = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '.';
    var jsGivenReportsDir = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : REPORTS_DESTINATION;

    var files = fs.readdirSync(`${jsGivenReportsDir}`).filter(filter);
    var scenarioReports = files.map(function (file) {
        return JSON.parse(fs.readFileSync(`${jsGivenReportsDir}/${file}`, 'utf-8'));
    });
    var groupNamesAndScenarios = Object.entries(_.groupBy(scenarioReports, function (_ref4) {
        var name = _ref4.groupReport.name;
        return name;
    }));
    var scenarioModels = groupNamesAndScenarios.map(function (_ref5) {
        var _ref6 = slicedToArray(_ref5, 2),
            groupName = _ref6[0],
            scenarioReports = _ref6[1];

        return toReportModel(groupName, scenarioReports);
    });

    var reportDir = `${reportPrefix}/jGiven-report`;

    fs.writeFileSync(`${reportDir}/data/metaData.js`, `jgivenReport.setMetaData({"created":"${new Date().toLocaleString()}","title":"J(s)Given Report","data":["data0.js"]} );\n`, 'utf-8');
    fs.writeFileSync(`${reportDir}/data/tags.js`, `jgivenReport.setTags({});\n`, 'utf-8');

    var json = JSON.stringify({ scenarios: scenarioModels });
    var buffer = zlib.gzipSync(Buffer.from(json, 'utf-8'));
    var base64 = buffer.toString('base64');
    fs.writeFileSync(`${reportDir}/data/data0.js`, `jgivenReport.addZippedScenarios('${base64}');\n`, 'utf-8');
}

function removeDir(dir) {
    return new Promise(function (resolve, reject) {
        rimraf(dir, function (error) {
            if (error) {
                reject(error);
            } else {
                resolve();
            }
        });
    });
}

function toReportModel(groupName, scenarioReports) {
    return {
        className: groupName,
        description: groupName,
        name: groupName,
        scenarios: scenarioReports.map(toScenarioModel),
        tagMap: {}
    };
}

function toScenarioModel(scenarioReport, index) {
    return {
        testMethodName: scenarioReport.name,
        tagIds: [],
        scenarioCases: scenarioReport.cases.map(function (scenarioCase, index) {
            return {
                caseNr: index + 1,
                derivedArguments: scenarioCase.args,
                description: '',
                durationInNanos: scenarioCase.durationInNanos,
                errorMessage: scenarioCase.errorMessage,
                explicitParameters: scenarioCase.args,
                explicitArguments: scenarioCase.args,
                stackTrace: scenarioCase.stackTrace,
                success: scenarioCase.successful,
                steps: _.flatMap(scenarioCase.parts, toSteps)
            };
        }),
        casesAsTable: scenarioReport.cases.length > 1,
        className: scenarioReport.groupReport.name,
        derivedParameters: scenarioReport.argumentNames,
        description: scenarioReport.name,
        durationInNanos: scenarioReport.cases.map(function (c) {
            return c.durationInNanos;
        }).reduce(function (a, b) {
            return a + b;
        }, 0),
        executionStatus: scenarioReport.executionStatus,
        explicitParameters: scenarioReport.argumentNames,
        extendedDescription: ''
    };
}

function toSteps(scenarioPart) {
    return scenarioPart.steps.map(function (step) {
        return {
            attachment: null,
            durationInNanos: step.durationInNanos,
            extendedDescription: '',
            isSectionTitle: false,
            name: step.methodName,
            nestedSteps: [],
            status: step.status,
            words: step.words.map(function (word) {
                return _extends({}, word.isIntroWord ? { isIntroWord: true } : {}, {
                    value: word.value
                }, word.scenarioParameterName !== null ? {
                    argumentInfo: {
                        argumentName: word.scenarioParameterName,
                        parameterName: word.scenarioParameterName,
                        formattedValue: word.value
                    }
                } : {});
            })
        };
    });
}

var _yargs$usage$command$ = yargs.usage('Usage: $0 <command>').command('report', 'Generate the reports', function (yars) {
    return yargs.option('fail', {
        describe: 'Generates the reports but return -1'
    });
}).command('clean', 'Remove the reports and intermediary files').demandCommand(1);
var argv = _yargs$usage$command$.argv;

var _argv$_ = slicedToArray(argv._, 1);
var command = _argv$_[0];

if (command === 'report') {
    var fail = !!argv.fail;
    generateJGivenReport(fail);
}

if (command === 'clean') {
    cleanReports();
}
