'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _ = _interopDefault(require('lodash'));
var retrieveArguments = _interopDefault(require('retrieve-arguments'));
var humanize = _interopDefault(require('string-humanize'));
var stripAnsi = _interopDefault(require('strip-ansi'));
var fs = _interopDefault(require('fs'));
var crypto = _interopDefault(require('crypto'));

var asyncGenerator = function () {
  function AwaitValue(value) {
    this.value = value;
  }

  function AsyncGenerator(gen) {
    var front, back;

    function send(key, arg) {
      return new Promise(function (resolve, reject) {
        var request = {
          key: key,
          arg: arg,
          resolve: resolve,
          reject: reject,
          next: null
        };

        if (back) {
          back = back.next = request;
        } else {
          front = back = request;
          resume(key, arg);
        }
      });
    }

    function resume(key, arg) {
      try {
        var result = gen[key](arg);
        var value = result.value;

        if (value instanceof AwaitValue) {
          Promise.resolve(value.value).then(function (arg) {
            resume("next", arg);
          }, function (arg) {
            resume("throw", arg);
          });
        } else {
          settle(result.done ? "return" : "normal", result.value);
        }
      } catch (err) {
        settle("throw", err);
      }
    }

    function settle(type, value) {
      switch (type) {
        case "return":
          front.resolve({
            value: value,
            done: true
          });
          break;

        case "throw":
          front.reject(value);
          break;

        default:
          front.resolve({
            value: value,
            done: false
          });
          break;
      }

      front = front.next;

      if (front) {
        resume(front.key, front.arg);
      } else {
        back = null;
      }
    }

    this._invoke = send;

    if (typeof gen.return !== "function") {
      this.return = undefined;
    }
  }

  if (typeof Symbol === "function" && Symbol.asyncIterator) {
    AsyncGenerator.prototype[Symbol.asyncIterator] = function () {
      return this;
    };
  }

  AsyncGenerator.prototype.next = function (arg) {
    return this._invoke("next", arg);
  };

  AsyncGenerator.prototype.throw = function (arg) {
    return this._invoke("throw", arg);
  };

  AsyncGenerator.prototype.return = function (arg) {
    return this._invoke("return", arg);
  };

  return {
    wrap: function (fn) {
      return function () {
        return new AsyncGenerator(fn.apply(this, arguments));
      };
    },
    await: function (value) {
      return new AwaitValue(value);
    }
  };
}();



var asyncToGenerator = function (fn) {
  return function () {
    var gen = fn.apply(this, arguments);
    return new Promise(function (resolve, reject) {
      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          return Promise.resolve(value).then(function (value) {
            step("next", value);
          }, function (err) {
            step("throw", err);
          });
        }
      }

      return step("next");
    });
  };
};

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();







var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};



var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};











var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};





var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();











var toArray = function (arr) {
  return Array.isArray(arr) ? arr : Array.from(arr);
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var asyncActionsSingleton = [];
function executeStepAndCollectAsyncActions(stepActionThatMayCallDoAsync) {
    asyncActionsSingleton = [];
    var collectedAsyncActions = [];

    try {
        stepActionThatMayCallDoAsync();
    } finally {
        collectedAsyncActions.push.apply(collectedAsyncActions, toConsumableArray(asyncActionsSingleton));
        asyncActionsSingleton = [];
    }

    return collectedAsyncActions;
}

function doAsync(action) {
    asyncActionsSingleton.push(action);
}

var Stage = function () {
    function Stage() {
        classCallCheck(this, Stage);
    }

    createClass(Stage, [{
        key: "and",
        value: function and() {
            return this;
        }
    }, {
        key: "but",
        value: function but() {
            return this;
        }
    }, {
        key: "with",
        value: function _with() {
            return this;
        }
    }, {
        key: "given",
        value: function given() {
            return this;
        }
    }, {
        key: "when",
        value: function when() {
            return this;
        }
    }, {
        key: "then",
        value: function then() {
            return this;
        }
    }]);
    return Stage;
}();

function getStageMetadataStoreProvider(key) {
    var KEY = `__JSGIVEN__INTERNAL__${key}__KEY__`;

    function getOrBuildStore(target) {
        if (!target[KEY]) {
            target[KEY] = new StageMetadataStoreImpl();
        }

        return target[KEY];
    }

    return {
        getStoreFromTarget(target) {
            return getOrBuildStore(target);
        },
        getStoreFromStageClass(stageClass) {
            return getOrBuildStore(stageClass.prototype);
        }
    };
}

var StageMetadataStoreImpl = function () {
    function StageMetadataStoreImpl() {
        classCallCheck(this, StageMetadataStoreImpl);
        this.properties = [];
    }

    createClass(StageMetadataStoreImpl, [{
        key: 'addProperty',
        value: function addProperty(property) {
            this.properties.push(property);
        }
    }, {
        key: 'getProperties',
        value: function getProperties() {
            return this.properties;
        }
    }]);
    return StageMetadataStoreImpl;
}();

function checkIsFunction(target, errorMessage) {
    if (!target || !_.isFunction(target)) {
        throw new Error(errorMessage);
    }
}

function checkIsParameter(func, parameterName, errorMessage) {
    if (!retrieveArguments(func).includes(parameterName)) {
        throw new Error(errorMessage);
    }
}

var storeProvider = getStageMetadataStoreProvider('@Hidden');

function Hidden(target, key, descriptor) {
    checkIsFunction(target[key], `@Hidden decorator can only be applied to methods: '${key}' is not a method.`);
    storeProvider.getStoreFromTarget(target).addProperty(key);
    return _extends({}, descriptor, { writable: true });
}
Hidden.addHiddenStep = function (stageClass, property) {
    checkIsFunction(
    // $FlowIgnore
    stageClass.prototype[property], `Hidden.addHiddenStep() can only be applied to methods: '${property}' is not a method.`);
    storeProvider.getStoreFromStageClass(stageClass).addProperty(property);
};

function isHiddenStep(stage, stepName) {
    var steps = storeProvider.getStoreFromTarget(stage).getProperties();
    return steps.includes(stepName);
}

function parametrized(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}

function parametrized1(parameters, func) {
    return {
        parameters: parameters.map(function (param) {
            return [param];
        }),
        func
    };
}
function parametrized2(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized3(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized4(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized5(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized6(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized7(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}

function wrapParameter(value, scenarioParameterName) {
    return {
        scenarioParameterName,
        value,
        IS_JSGIVEN_WRAPPER_PARAMETER: true
    };
}

function decodeParameter(parameter, stepParameterName, formatters) {
    if (parameter instanceof Object && parameter.IS_JSGIVEN_WRAPPER_PARAMETER) {
        var wrapped = parameter;
        return _extends({}, wrapped, { stepParameterName, formatters });
    } else {
        if (stepParameterName === undefined) {
            throw new Error('cant be undefined');
        }
        return {
            value: parameter,
            scenarioParameterName: null,
            stepParameterName,
            formatters
        };
    }
}

var storeProvider$1 = getStageMetadataStoreProvider('@ParameterFormatters');

function getFormatters(stage, requestedStepMethodName, requestedParameterName) {
    var formattings = storeProvider$1.getStoreFromTarget(stage).getProperties();
    return formattings.filter(function (_ref) {
        var stepMethodName = _ref.stepMethodName,
            parameterName = _ref.parameterName;
        return requestedStepMethodName === stepMethodName && requestedParameterName === parameterName;
    }).map(function (_ref2) {
        var formatter = _ref2.formatter;
        return formatter;
    });
}

function restParameterName() {
    return 'JSGIVEN_REST_PARAMETER_NAME';
}

function buildParameterFormatter(formatter) {
    var parameterFormatter = function parameterFormatter() {
        for (var _len = arguments.length, parameterNames = Array(_len), _key = 0; _key < _len; _key++) {
            parameterNames[_key] = arguments[_key];
        }

        var decorator = function decorator(target, stepMethodName, descriptor) {
            parameterNames.forEach(function (parameterName) {
                checkIsFunction(target[stepMethodName], `Formatter decorators can only be applied to methods: '${stepMethodName}' is not a method.`);
                checkIsParameter(target[stepMethodName], parameterName, `Formatter decorator cannot be applied on method: ${stepMethodName}(): parameter '${parameterName}' was not found.`);
                storeProvider$1.getStoreFromTarget(target).addProperty({
                    formatter,
                    parameterName,
                    stepMethodName
                });
            });
            return _extends({}, descriptor, { writable: true });
        };
        return decorator;
    };
    parameterFormatter.formatParameter = function (stageClass, stepMethodName) {
        for (var _len2 = arguments.length, parameterNames = Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
            parameterNames[_key2 - 2] = arguments[_key2];
        }

        parameterNames.forEach(function (parameterName) {
            checkIsFunction(
            // $FlowIgnore
            stageClass.prototype[stepMethodName], `Formatter.formatParameter() can only be applied to methods: '${stepMethodName}' is not a method.`);
            checkIsParameter(
            // $FlowIgnore
            stageClass.prototype[stepMethodName], parameterName, `Formatter.formatParameter() cannot be applied on method: ${stepMethodName}(): parameter '${parameterName}' was not found.`);
            storeProvider$1.getStoreFromStageClass(stageClass).addProperty({
                formatter,
                parameterName,
                stepMethodName
            });
        });
    };

    return parameterFormatter;
}

var Quoted = buildParameterFormatter(function (parameterValue) {
    return `"${parameterValue}"`;
});

var QuotedWith = function QuotedWith(quoteCharacter) {
    return buildParameterFormatter(function (parameterValue) {
        return `${quoteCharacter}${parameterValue}${quoteCharacter}`;
    });
};

var NotFormatter = buildParameterFormatter(function (parameterValue) {
    return parameterValue ? '' : 'not';
});

var REPORTS_DESTINATION = '.jsGiven-reports';

var INTRO_WORD_METHODS = ['given', 'when', 'then', 'and', 'but', 'with'];

var ScenarioPart = function () {
    function ScenarioPart(kind) {
        var steps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
        classCallCheck(this, ScenarioPart);

        this.kind = kind;
        this.steps = steps;
        this.introWord = null;
    }

    createClass(ScenarioPart, [{
        key: 'stageMethodCalled',
        value: function stageMethodCalled(methodName, parameters, stepStatus, executionTimeInNanos) {
            if (INTRO_WORD_METHODS.find(function (introWord) {
                return introWord === methodName;
            })) {
                this.introWord = methodName;
            } else {
                var isFirstStep = this.steps.length === 0;
                this.steps.push(new Step(methodName, parameters, isFirstStep, this.introWord, stepStatus, executionTimeInNanos));
                this.introWord = null;
            }
        }
    }]);
    return ScenarioPart;
}();

var Step = function Step(methodName, parameters, isFirstStep, introWord, stepStatus, durationInNanos) {
    classCallCheck(this, Step);

    var TWO_DOLLAR_PLACEHOLDER = 'zzblablaescapedollarsignplaceholdertpolm';

    var parametersCopy = [].concat(toConsumableArray(parameters));

    var words = [].concat(toConsumableArray(methodName // 'a_bill_of_$_$$'
    .replace('$$', TWO_DOLLAR_PLACEHOLDER) // 'a_bill_of_$_TWO_DOLLAR_PLACEHOLDER'
    .split('$') // ['a_bill_of', 'TWO_DOLLAR_PLACEHOLDER']
    .map(function (word) {
        return _.lowerCase(humanize(word));
    }) //  ['a bill of', 'TWO_DOLLAR_PLACEHOLDER']
    .reduce(function (previous, newString, index) {
        if (index === 0) {
            return [{ word: newString, scenarioParameterName: null }];
        }

        var formattedParameters = void 0;
        if (parametersCopy.length > 0) {
            var _parametersCopy$splic = parametersCopy.splice(0, 1),
                _parametersCopy$splic2 = slicedToArray(_parametersCopy$splic, 1),
                parameter = _parametersCopy$splic2[0];

            var _word = formatParameter(parameter.value, parameter.formatters);
            formattedParameters = [{
                word: _word,
                scenarioParameterName: parameter.scenarioParameterName
            }];
        } else {
            formattedParameters = [];
        }

        return [].concat(toConsumableArray(previous), toConsumableArray(formattedParameters), [{ word: newString, scenarioParameterName: null }]);
    }, []) //  ['a bill of', '500', 'TWO_DOLLAR_PLACEHOLDER']
    .filter(function (_ref) {
        var word = _ref.word;
        return word !== '';
    }) // If one puts a $ at the end of the method, this adds a useless '' at the end
    .map(function (_ref2) {
        var word = _ref2.word,
            scenarioParameterName = _ref2.scenarioParameterName;
        return {
            word: word.replace(TWO_DOLLAR_PLACEHOLDER, '$'),
            scenarioParameterName
        };
    }) //  ['a bill of', '500', '$']
    .map(toWord)), toConsumableArray(parametersCopy.map(function (parameter) {
        return {
            word: formatParameter(parameter.value, parameter.formatters),
            scenarioParameterName: parameter.scenarioParameterName
        };
    }).map(toWord)));

    if (introWord !== null) {
        words = [toIntroWord(introWord)].concat(toConsumableArray(words));
    }

    if (isFirstStep) {
        var _words = words,
            _words2 = toArray(_words),
            _words2$ = _words2[0],
            value = _words2$.value,
            isIntroWord = _words2$.isIntroWord,
            rest = _words2.slice(1);

        words = [new Word(_.upperFirst(value), isIntroWord)].concat(toConsumableArray(rest));
    }
    this.words = words;
    this.name = words.map(function (_ref3) {
        var value = _ref3.value;
        return value;
    }).join(' ');
    this.status = stepStatus;
    this.durationInNanos = durationInNanos;

    function toWord(_ref4) {
        var word = _ref4.word,
            scenarioParameterName = _ref4.scenarioParameterName;

        return new Word(word, false, scenarioParameterName);
    }

    function toIntroWord(value) {
        return new Word(value, true, null);
    }
};

var Word = function Word(value, isIntroWord) {
    var scenarioParameterName = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    classCallCheck(this, Word);

    this.value = value;
    this.isIntroWord = isIntroWord;
    this.scenarioParameterName = scenarioParameterName;
};

var ScenarioCase = function ScenarioCase() {
    var args = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    var parts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    classCallCheck(this, ScenarioCase);

    this.args = args;
    this.parts = parts;
};

var ScenarioReport = function () {
    function ScenarioReport(groupReport, name, cases, argumentNames) {
        classCallCheck(this, ScenarioReport);

        this.groupReport = groupReport;
        this.name = name;
        this.cases = cases;
        this.argumentNames = argumentNames;
    }

    createClass(ScenarioReport, [{
        key: 'dumpToFile',
        value: function dumpToFile(reportsDestination) {
            createDirOrDoNothingIfExists(reportsDestination);
            var fileName = computeScenarioFileName(this.groupReport.name, this.name);
            fs.writeFileSync(`${reportsDestination}/${fileName}`, JSON.stringify(this), 'utf-8');
        }
    }]);
    return ScenarioReport;
}();

function formatParameter(parameter, formatters) {
    if (formatters.length > 0) {
        var value = parameter;
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = formatters[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var formatter = _step.value;

                value = formatter(value);
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        return formatParameter(value, []);
    }
    if (_.isString(parameter)) {
        return parameter;
    } else if (_.isObject(parameter) || Array.isArray(parameter)) {
        if (parameter.toString && parameter.toString !== Object.prototype.toString && parameter.toString !== Array.prototype.toString) {
            return parameter.toString();
        }
        return JSON.stringify(parameter);
    } else {
        return parameter && parameter.toString ? parameter.toString() : JSON.stringify(parameter);
    }
}

var GroupReport = function GroupReport(name) {
    classCallCheck(this, GroupReport);

    this.name = name;
};

function createDirOrDoNothingIfExists(path) {
    try {
        fs.mkdirSync(path);
    } catch (error) {
        if (error.code !== 'EEXIST') {
            throw error;
        } else {
            // do nothing
        }
    }
}

function computeScenarioFileName(groupName, scenarioName) {
    var hash = crypto.createHash('sha256');
    hash.update(groupName + '\n' + scenarioName);
    return hash.digest('hex');
}

var stateProvider = getStageMetadataStoreProvider('@State');

function State(target, key, descriptor) {
    stateProvider.getStoreFromTarget(target).addProperty(key);
    return _extends({}, descriptor, { writable: true });
}
State.addProperty = function (stageClass, property) {
    stateProvider.getStoreFromStageClass(stageClass).addProperty(property);
};

function copyStateToOtherStages(originalStage, allStages) {
    allStages.forEach(function (targetStage) {
        if (originalStage !== targetStage) {
            copyStateProperties(originalStage, targetStage);
        }
    });
}

function copyStateProperties(source, target) {
    if (source && target) {
        var propertyNames = _.intersection(stateProvider.getStoreFromTarget(source).getProperties(), stateProvider.getStoreFromTarget(target).getProperties());
        propertyNames.forEach(function (propertyName) {
            // Need to convert to any to avoid typechecking
            var sourceAny = source;
            var targetAny = target;
            targetAny[propertyName] = sourceAny[propertyName];
        });
    }
}

var Timer = function () {
    function Timer() {
        classCallCheck(this, Timer);

        this.startTime = process.hrtime();
    }

    createClass(Timer, [{
        key: "elapsedTimeInNanoseconds",
        value: function elapsedTimeInNanoseconds() {
            var NS_PER_SEC = 1e9;

            var _process$hrtime = process.hrtime(this.startTime),
                _process$hrtime2 = slicedToArray(_process$hrtime, 2),
                seconds = _process$hrtime2[0],
                nanoseconds = _process$hrtime2[1];

            return seconds * NS_PER_SEC + nanoseconds;
        }
    }]);
    return Timer;
}();

function scenario(options, scenarioFunction) {
    var scenarioOptions = _extends({
        tags: []
    }, options);
    return {
        scenarioFunction,
        tags: scenarioOptions.tags
    };
}

var ScenarioRunner = function () {
    function ScenarioRunner() {
        var reportsDestination = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : REPORTS_DESTINATION;
        classCallCheck(this, ScenarioRunner);

        this.reportsDestination = reportsDestination;
    }

    createClass(ScenarioRunner, [{
        key: 'setup',
        value: function setup(groupFunc, testFunc) {
            this.groupFunc = groupFunc;
            this.testFunc = testFunc;
        }
    }, {
        key: 'scenarios',
        value: function scenarios(groupName, stagesParams, scenariosDescriptions) {
            var _this = this;

            if (!this.groupFunc || !this.testFunc) {
                throw new Error('JsGiven is not initialized, please call setupForRspec() or setupForAva() in your test code');
            }

            var report = new GroupReport(groupName);

            var currentStages = undefined;

            var stageBuilder = void 0;

            if (Array.isArray(stagesParams)) {
                var self = this;

                var _stagesParams = slicedToArray(stagesParams, 3),
                    givenClass = _stagesParams[0],
                    whenClass = _stagesParams[1],
                    thenClass = _stagesParams[2];

                stageBuilder = function stageBuilder(runningScenario) {
                    var givenStage = self.buildStage(givenClass, runningScenario);
                    var whenStage = self.buildStage(whenClass, runningScenario);
                    var thenStage = self.buildStage(thenClass, runningScenario);
                    return { givenStage, whenStage, thenStage };
                };
            } else {
                var _self = this;
                var _givenClass = stagesParams;

                stageBuilder = function stageBuilder(runningScenario) {
                    var givenStage = _self.buildStage(_givenClass, runningScenario);
                    var whenStage = givenStage;
                    var thenStage = givenStage;
                    return { givenStage, whenStage, thenStage };
                };
            }

            var scenariosParam = {
                given: function given() {
                    if (currentStages) {
                        return currentStages.givenStage.given();
                    } else {
                        throw new Error('given() may only be called in scenario');
                    }
                },
                when: function when() {
                    if (currentStages) {
                        return currentStages.whenStage.when();
                    } else {
                        throw new Error('when() may only be called in scenario');
                    }
                },
                then: function then() {
                    if (currentStages) {
                        return currentStages.thenStage.then();
                    } else {
                        throw new Error('then() may only be called in scenario');
                    }
                }
            };

            this.groupFunc(groupName, function () {
                var scenarios = scenariosDescriptions(scenariosParam);

                getScenarios(scenarios).forEach(function (_ref) {
                    var scenarioPropertyName = _ref.scenarioPropertyName,
                        cases = _ref.cases,
                        argumentNames = _ref.argumentNames;

                    var scenarioNameForHumans = humanize(scenarioPropertyName);
                    var scenario = _this.addScenario(report, scenarioNameForHumans, argumentNames);

                    var casesCount = 0;
                    cases.forEach(function (_ref2, index) {
                        var caseFunction = _ref2.caseFunction,
                            wrappedArgs = _ref2.wrappedArgs;

                        var caseDescription = cases.length === 1 ? scenarioNameForHumans : `${scenarioNameForHumans} #${index + 1}`;
                        _this.testFunc(caseDescription, function () {
                            var continueAsyncWork = function () {
                                var _ref3 = asyncToGenerator(function* (self, asyncActions, stage, steps, i, initialStepTimer) {
                                    var executeAsyncActions = function () {
                                        var _ref4 = asyncToGenerator(function* (asyncActions) {
                                            var _iteratorNormalCompletion = true;
                                            var _didIteratorError = false;
                                            var _iteratorError = undefined;

                                            try {
                                                for (var _iterator = asyncActions[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                                                    var asyncAction = _step.value;

                                                    yield asyncAction();
                                                }
                                            } catch (err) {
                                                _didIteratorError = true;
                                                _iteratorError = err;
                                            } finally {
                                                try {
                                                    if (!_iteratorNormalCompletion && _iterator.return) {
                                                        _iterator.return();
                                                    }
                                                } finally {
                                                    if (_didIteratorError) {
                                                        throw _iteratorError;
                                                    }
                                                }
                                            }
                                        });

                                        return function executeAsyncActions(_x8) {
                                            return _ref4.apply(this, arguments);
                                        };
                                    }();

                                    try {
                                        // Execute async actions for current step
                                        try {
                                            yield executeAsyncActions(asyncActions);
                                        } catch (error) {
                                            caughtError = error;
                                            var _stepActions2 = steps[i].stepActions;

                                            _stepActions2.markStepAsFailed(initialStepTimer);
                                        }
                                        if (!caughtError) {
                                            var _stepActions3 = steps[i].stepActions;

                                            _stepActions3.markStepAsPassed(initialStepTimer);
                                            copyStateToOtherStages(stage, runningScenario.stages);
                                        }

                                        // Execute further steps and their async actions
                                        for (var j = i + 1; j < steps.length; j++) {
                                            var _steps$j = steps[j],
                                                _stepActions4 = _steps$j.stepActions,
                                                _stage2 = _steps$j.stage;

                                            var _stepTimer = new Timer();
                                            if (!caughtError) {
                                                try {
                                                    var actions = executeStepAndCollectAsyncActions(_stepActions4.executeStep);
                                                    yield executeAsyncActions(actions);
                                                } catch (error) {
                                                    caughtError = error;
                                                    _stepActions4.markStepAsFailed(_stepTimer);
                                                }

                                                if (!caughtError) {
                                                    _stepActions4.markStepAsPassed(_stepTimer);
                                                    copyStateToOtherStages(_stage2, runningScenario.stages);
                                                }
                                            } else {
                                                _stepActions4.markStepAsSkipped(_stepTimer);
                                            }
                                        }
                                    } finally {
                                        caseCompleted(self);
                                    }

                                    if (caughtError) {
                                        throw caughtError;
                                    }
                                });

                                return function continueAsyncWork(_x2, _x3, _x4, _x5, _x6, _x7) {
                                    return _ref3.apply(this, arguments);
                                };
                            }();

                            var caughtError = null;
                            var runningScenario = {
                                state: 'COLLECTING_STEPS',
                                stages: [],
                                steps: [],
                                caseArguments: wrappedArgs,
                                formattedCaseArguments: {}
                            };

                            _this.beginCase(scenario);

                            // Build stages
                            currentStages = stageBuilder(runningScenario);

                            // Collecting steps
                            caseFunction();

                            _this.setCaseArguments(runningScenario.caseArguments, runningScenario.formattedCaseArguments);

                            // Execute scenario
                            runningScenario.state = 'RUNNING';
                            var runningSynchronously = true;
                            try {
                                var _steps = runningScenario.steps;

                                for (var i = 0; i < _steps.length; i++) {
                                    var _steps$i = _steps[i],
                                        _stepActions = _steps$i.stepActions,
                                        _stage = _steps$i.stage;

                                    var stepTimer = new Timer();
                                    var asyncActions = [];

                                    if (!caughtError) {
                                        try {
                                            asyncActions = executeStepAndCollectAsyncActions(_stepActions.executeStep);
                                        } catch (error) {
                                            caughtError = error;
                                            _stepActions.markStepAsFailed(stepTimer);
                                        }
                                    } else {
                                        _stepActions.markStepAsSkipped(stepTimer);
                                    }

                                    if (!caughtError) {
                                        if (asyncActions.length === 0) {
                                            _stepActions.markStepAsPassed(stepTimer);
                                            copyStateToOtherStages(_stage, runningScenario.stages);
                                        } else {
                                            // If we have asynchronous actions, we
                                            // switch to asynchronous mode
                                            // We return immediately a promise and
                                            // the execution is now asynchronous
                                            // This ensures that scenarios that do
                                            // not require asynchronous processing
                                            // are still executed synchronously
                                            runningSynchronously = false;
                                            return continueAsyncWork(_this, asyncActions, _stage, _steps, i, stepTimer);
                                        }
                                    }
                                }
                            } finally {
                                if (runningSynchronously) {
                                    caseCompleted(_this);
                                }
                            }

                            if (caughtError) {
                                throw caughtError;
                            }

                            function caseCompleted(self) {
                                if (caughtError) {
                                    self.caseFailed(caughtError);
                                } else {
                                    self.caseSucceeded();
                                }
                                casesCount++;
                                if (casesCount === cases.length) {
                                    self.scenarioCompleted(scenario);
                                }
                                currentStages = undefined;
                            }
                        });
                    });
                });

                function getScenarios(scenarios) {
                    var scenarioDescriptions = Object.keys(scenarios).map(function (scenarioPropertyName) {
                        var scenarioDescription = scenarios[scenarioPropertyName];
                        var scenarioFunction = scenarioDescription.scenarioFunction;

                        if (scenarioFunction instanceof Function) {
                            return {
                                scenarioPropertyName,
                                cases: [{
                                    caseFunction: scenarioFunction,
                                    wrappedArgs: []
                                }],
                                argumentNames: []
                            };
                        } else {
                            var _ref5 = scenarioFunction,
                                parameters = _ref5.parameters,
                                func = _ref5.func;

                            var _argumentNames = retrieveArguments(func);

                            return {
                                scenarioPropertyName,
                                cases: parameters.map(function (parametersForCase, z) {
                                    var parametersForTestFunction = parametersForCase.map(function (parameter, index) {
                                        return wrapParameter(parameter, _argumentNames[index]);
                                    });
                                    return {
                                        caseFunction: function caseFunction() {
                                            return func.apply(undefined, toConsumableArray(parametersForTestFunction));
                                        },
                                        wrappedArgs: parametersForTestFunction
                                    };
                                }),
                                argumentNames: _argumentNames
                            };
                        }
                    });
                    return scenarioDescriptions;
                }
            });
        }
    }, {
        key: 'addScenario',
        value: function addScenario(report, scenarioNameForHumans, argumentNames) {
            return new ScenarioReport(report, scenarioNameForHumans, [], argumentNames);
        }
    }, {
        key: 'beginCase',
        value: function beginCase(scenario) {
            var currentCase = new ScenarioCase();
            this.currentCase = currentCase;
            scenario.cases.push(currentCase);
            this.currentCaseTimer = new Timer();
        }
    }, {
        key: 'setCaseArguments',
        value: function setCaseArguments(caseArguments, formattedCaseArguments) {
            this.currentCase.args = caseArguments.map(function (wp) {
                return formattedCaseArguments[wp.scenarioParameterName];
            });
        }
    }, {
        key: 'caseFailed',
        value: function caseFailed(error) {
            var durationInNanos = this.currentCaseTimer.elapsedTimeInNanoseconds();
            this.currentCase.successful = false;
            this.currentCase.durationInNanos = durationInNanos;
            this.currentCase.errorMessage = stripAnsi(error.message);
            this.currentCase.stackTrace = stripAnsi(error.stack).split('\n');
        }
    }, {
        key: 'caseSucceeded',
        value: function caseSucceeded() {
            var durationInNanos = this.currentCaseTimer.elapsedTimeInNanoseconds();
            this.currentCase.successful = true;
            this.currentCase.durationInNanos = durationInNanos;
        }
    }, {
        key: 'addGivenPart',
        value: function addGivenPart() {
            this.currentPart = new ScenarioPart('GIVEN');
            this.currentCase.parts.push(this.currentPart);
        }
    }, {
        key: 'addWhenPart',
        value: function addWhenPart() {
            this.currentPart = new ScenarioPart('WHEN');
            this.currentCase.parts.push(this.currentPart);
        }
    }, {
        key: 'addThenPart',
        value: function addThenPart() {
            this.currentPart = new ScenarioPart('THEN');
            this.currentCase.parts.push(this.currentPart);
        }
    }, {
        key: 'stepPassed',
        value: function stepPassed(methodName, decodedParameters, stepTimer) {
            this.currentPart.stageMethodCalled(methodName, decodedParameters, 'PASSED', stepTimer.elapsedTimeInNanoseconds());
        }
    }, {
        key: 'stepFailed',
        value: function stepFailed(methodName, decodedParameters, stepTimer) {
            this.currentPart.stageMethodCalled(methodName, decodedParameters, 'FAILED', stepTimer.elapsedTimeInNanoseconds());
        }
    }, {
        key: 'stepSkipped',
        value: function stepSkipped(methodName, decodedParameters, stepTimer) {
            this.currentPart.stageMethodCalled(methodName, decodedParameters, 'SKIPPED', stepTimer.elapsedTimeInNanoseconds());
        }
    }, {
        key: 'scenarioCompleted',
        value: function scenarioCompleted(scenario) {
            var executionStatus = scenario.cases.some(function (c) {
                return !c.successful;
            }) ? 'FAILED' : 'SUCCESS';
            scenario.executionStatus = executionStatus;
            scenario.dumpToFile(this.reportsDestination);
        }
    }, {
        key: 'buildStage',
        value: function buildStage(tClass, runningScenario) {
            var _this3 = this;

            // $FlowIgnore
            var extendedClass = function (_tClass) {
                inherits(extendedClass, _tClass);

                function extendedClass() {
                    classCallCheck(this, extendedClass);
                    return possibleConstructorReturn(this, (extendedClass.__proto__ || Object.getPrototypeOf(extendedClass)).apply(this, arguments));
                }

                return extendedClass;
            }(tClass);

            // Flowtype really can't type this constructor invocation
            // Therefore we have to cast it as any :(


            var instance = new extendedClass();

            var extendedPrototype = Object.getPrototypeOf(instance);
            var classPrototype = Object.getPrototypeOf(extendedPrototype);

            var stages = runningScenario.stages;


            getAllMethods(classPrototype).forEach(function (methodName) {
                var self = _this3;

                extendedPrototype[methodName] = function () {
                    var _this4 = this;

                    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
                        args[_key] = arguments[_key];
                    }

                    var state = runningScenario.state,
                        steps = runningScenario.steps,
                        caseArguments = runningScenario.caseArguments,
                        formattedCaseArguments = runningScenario.formattedCaseArguments;


                    var stepParameterNames = retrieveArguments(classPrototype[methodName]);
                    var decodedParameters = args.map(function (arg, index) {
                        var parameterName = index < stepParameterNames.length ? stepParameterNames[index] : restParameterName();
                        return decodeParameter(arg, parameterName, getFormatters(instance, methodName, parameterName));
                    });

                    caseArguments.forEach(function (_ref6) {
                        var scenarioParameterName = _ref6.scenarioParameterName,
                            value = _ref6.value;

                        if (formattedCaseArguments[scenarioParameterName] === undefined) {
                            var foundDecodedParameter = decodedParameters.find(function (dp) {
                                return dp.scenarioParameterName === scenarioParameterName;
                            });
                            if (foundDecodedParameter) {
                                // eslint-disable-next-line standard/computed-property-even-spacing
                                formattedCaseArguments[scenarioParameterName] = formatParameter(value, foundDecodedParameter.formatters);
                            }
                        }
                    });

                    switch (state) {
                        case 'COLLECTING_STEPS':
                            {
                                steps.push({
                                    stepActions: {
                                        executeStep: function executeStep() {
                                            return extendedPrototype[methodName].apply(_this4, args);
                                        },
                                        markStepAsPassed: function markStepAsPassed(stepTimer) {
                                            if (!isHiddenStep(_this4, methodName)) {
                                                self.stepPassed(methodName, decodedParameters, stepTimer);
                                            }
                                        },
                                        markStepAsFailed: function markStepAsFailed(stepTimer) {
                                            if (!isHiddenStep(_this4, methodName)) {
                                                self.stepFailed(methodName, decodedParameters, stepTimer);
                                            }
                                        },
                                        markStepAsSkipped: function markStepAsSkipped(stepTimer) {
                                            if (!isHiddenStep(_this4, methodName)) {
                                                insertNewPartIfRequired();
                                                self.stepSkipped(methodName, decodedParameters, stepTimer);
                                            }
                                        }
                                    },
                                    stage: this
                                });
                                return this;
                            }
                        default:
                            {
                                // eslint-disable-next-line no-unused-vars
                                insertNewPartIfRequired();

                                // Pass the real arguments instead of the wrapped values
                                var values = decodedParameters.map(function (decodedParameter) {
                                    return decodedParameter.value;
                                });
                                var result = classPrototype[methodName].apply(this, values);

                                return result;
                            }
                    }
                };

                function insertNewPartIfRequired() {
                    if (methodName === 'given') {
                        self.addGivenPart();
                    }
                    if (methodName === 'when') {
                        self.addWhenPart();
                    }
                    if (methodName === 'then') {
                        self.addThenPart();
                    }
                }
            });

            stages.push(instance);

            return instance;

            function getAllMethods(obj) {
                var allMethods = [];
                var current = obj;
                do {
                    var props = Object.getOwnPropertyNames(current);
                    props.forEach(function (prop) {
                        if (allMethods.indexOf(prop) === -1) {
                            if (_.isFunction(current[prop])) {
                                allMethods.push(prop);
                            }
                        }
                    });
                } while (current = Object.getPrototypeOf(current));

                return allMethods;
            }
        }
    }]);
    return ScenarioRunner;
}();

var INSTANCE = new ScenarioRunner();

function scenarios(groupName, stagesParam, scenarioFunc) {
    return INSTANCE.scenarios(groupName, stagesParam, scenarioFunc);
}

function setupForRspec(describe, it) {
    return INSTANCE.setup(function (groupName, suiteFunc) {
        describe(groupName, suiteFunc);
    }, function (testName, testFunc) {
        it(testName, testFunc);
    });
}

function setupForAva(test) {
    var capturedGroupName = '';
    return INSTANCE.setup(function (groupName, suiteFunc) {
        capturedGroupName = groupName;
        suiteFunc();
    }, function (testName, testFunc) {
        test(`${capturedGroupName} / ${testName}`, testFunc);
    });
}

function buildTag(tagName) {
    var _ref = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
        _ref$description = _ref.description,
        description = _ref$description === undefined ? '' : _ref$description,
        _ref$style = _ref.style,
        style = _ref$style === undefined ? '' : _ref$style,
        _ref$values = _ref.values,
        values = _ref$values === undefined ? [] : _ref$values,
        _ref$parentTags = _ref.parentTags,
        parentTags = _ref$parentTags === undefined ? [] : _ref$parentTags,
        _ref$prependName = _ref.prependName,
        prependName = _ref$prependName === undefined ? false : _ref$prependName,
        _ref$tagDescriptionGe = _ref.tagDescriptionGenerator,
        tagDescriptionGenerator = _ref$tagDescriptionGe === undefined ? function (_ref2) {
        var description = _ref2.description;
        return description;
    } : _ref$tagDescriptionGe;

    var tag = function tag() {
        for (var _len = arguments.length, values = Array(_len), _key = 0; _key < _len; _key++) {
            values[_key] = arguments[_key];
        }

        var buildTagOptions = {
            description,
            style,
            values,
            parentTags,
            prependName,
            tagDescriptionGenerator
        };
        return buildTag(tagName, buildTagOptions);
    };
    tag.tagName = tagName;
    tag.description = tag.description;
    tag.values = values;
    tag.parentTags = parentTags;
    tag.prependName = prependName;
    tag.tagDescriptionGenerator = tagDescriptionGenerator;

    return tag;
}

exports.doAsync = doAsync;
exports.Hidden = Hidden;
exports.parametrized = parametrized;
exports.parametrized1 = parametrized1;
exports.parametrized2 = parametrized2;
exports.parametrized3 = parametrized3;
exports.parametrized4 = parametrized4;
exports.parametrized5 = parametrized5;
exports.parametrized6 = parametrized6;
exports.parametrized7 = parametrized7;
exports.Quoted = Quoted;
exports.QuotedWith = QuotedWith;
exports.NotFormatter = NotFormatter;
exports.buildParameterFormatter = buildParameterFormatter;
exports.scenario = scenario;
exports.scenarios = scenarios;
exports.Stage = Stage;
exports.State = State;
exports.setupForRspec = setupForRspec;
exports.setupForAva = setupForAva;
exports.buildTag = buildTag;
